@@ -1,6 +1,7 @@
 from functools import partial, singledispatch
 
 import cgen as c
+import numpy as np
 
 from devito.core.operator import OperatorCore
 from devito.data import FULL
@@ -10,16 +11,16 @@ from devito.logger import warning
 from devito.mpi.routines import CopyBuffer, SendRecv, HaloUpdate
 from devito.passes.clusters import (Lift, cire, cse, eliminate_arrays, extract_increments,
                                     factorize, fuse, optimize_pows)
-from devito.passes.iet import (DataManager, Storage, Ompizer, ParallelIteration,
+from devito.passes.iet import (DataManager, Storage, Ompizer, OpenMPIteration,
                                ParallelTree, optimize_halospots, mpiize, hoist_prodders,
                                iet_pass)
-from devito.tools import as_tuple, filter_sorted, generator, timed_pass
+from devito.tools import as_tuple, filter_sorted, timed_pass
 
 __all__ = ['DeviceOpenMPNoopOperator', 'DeviceOpenMPOperator',
            'DeviceOpenMPCustomOperator']
 
 
-class DeviceOpenMPIteration(ParallelIteration):
+class DeviceOpenMPIteration(OpenMPIteration):
 
     @classmethod
     def _make_construct(cls, **kwargs):
@@ -33,16 +34,6 @@ class DeviceOpenMPIteration(ParallelIteration):
 
 class DeviceOmpizer(Ompizer):
 
-    COLLAPSE_NCORES = 1
-    """
-    Always collapse when possible.
-    """
-
-    COLLAPSE_WORK = 1
-    """
-    Always collapse when possible.
-    """
-
     lang = dict(Ompizer.lang)
     lang.update({
         'map-enter-to': lambda i, j:
@@ -124,7 +115,7 @@ class DeviceOmpizer(Ompizer):
 
         return root, partree, collapsed
 
-    def _make_parregion(self, partree):
+    def _make_parregion(self, partree, *args):
         # no-op for now
         return partree
 
@@ -141,24 +132,18 @@ class DeviceOpenMPDataManager(DataManager):
 
     _Parallelizer = DeviceOmpizer
 
-    def _alloc_array_on_high_bw_mem(self, obj, storage):
-        if obj in storage._high_bw_mem:
-            return
+    def _alloc_array_on_high_bw_mem(self, site, obj, storage):
+        _storage = Storage()
+        super()._alloc_array_on_high_bw_mem(site, obj, _storage)
 
-        super(DeviceOpenMPDataManager, self)._alloc_array_on_high_bw_mem(obj, storage)
-
-        decl, alloc, free = storage._high_bw_mem[obj]
-
-        alloc = c.Collection([alloc, self._Parallelizer._map_alloc(obj)])
-        free = c.Collection([self._Parallelizer._map_delete(obj), free])
-
-        storage._high_bw_mem[obj] = (decl, alloc, free)
-
-    def _map_function_on_high_bw_mem(self, obj, storage, read_only=False):
-        """Place a Function in the high bandwidth memory."""
-        if obj in storage._high_bw_mem:
-            return
+        allocs = _storage[site].allocs + [self._Parallelizer._map_alloc(obj)]
+        frees = [self._Parallelizer._map_delete(obj)] + _storage[site].frees
+        storage.update(obj, site, allocs=allocs, frees=frees)
 
+    def _map_function_on_high_bw_mem(self, site, obj, storage, read_only=False):
+        """
+        Place a Function in the high bandwidth memory.
+        """
         alloc = self._Parallelizer._map_to(obj)
 
         if read_only is False:
@@ -167,7 +152,7 @@ class DeviceOpenMPDataManager(DataManager):
         else:
             free = self._Parallelizer._map_delete(obj)
 
-        storage._high_bw_mem[obj] = (None, alloc, free)
+        storage.update(obj, site, allocs=alloc, frees=free)
 
     @iet_pass
     def place_ondevice(self, iet, **kwargs):
@@ -195,9 +180,9 @@ class DeviceOpenMPDataManager(DataManager):
             # Populate `storage`
             storage = Storage()
             for i in filter_sorted(writes):
-                self._map_function_on_high_bw_mem(i, storage)
+                self._map_function_on_high_bw_mem(iet, i, storage)
             for i in filter_sorted(reads):
-                self._map_function_on_high_bw_mem(i, storage, read_only=True)
+                self._map_function_on_high_bw_mem(iet, i, storage, read_only=True)
 
             iet = self._dump_storage(iet, storage)
 
@@ -243,24 +228,49 @@ class DeviceOpenMPNoopOperator(OperatorCore):
     Minimum operation count of a sum-of-product aliasing expression to be optimized away.
     """
 
+    PAR_CHUNK_NONAFFINE = 3
+    """
+    Coefficient to adjust the chunk size in non-affine parallel loops.
+    """
+
     @classmethod
     def _normalize_kwargs(cls, **kwargs):
-        options = kwargs['options']
+        o = {}
+        oo = kwargs['options']
+
+        # Execution modes
+        o['mpi'] = oo.pop('mpi')
 
         # Strictly unneccesary, but make it clear that this Operator *will*
         # generate OpenMP code, bypassing any `openmp=False` provided in
         # input to Operator
-        options.pop('openmp')
+        oo.pop('openmp')
 
-        options['cire-repeats'] = {
-            'invariants': options.pop('cire-repeats-inv') or cls.CIRE_REPEATS_INV,
-            'sops': options.pop('cire-repeats-sops') or cls.CIRE_REPEATS_SOPS
+        # CIRE
+        o['min-storage'] = False
+        o['cire-maxpar'] = oo.pop('cire-maxpar', True)
+        o['cire-repeats'] = {
+            'invariants': oo.pop('cire-repeats-inv', cls.CIRE_REPEATS_INV),
+            'sops': oo.pop('cire-repeats-sops', cls.CIRE_REPEATS_SOPS)
         }
-        options['cire-mincost'] = {
-            'invariants': options.pop('cire-mincost-inv') or cls.CIRE_MINCOST_INV,
-            'sops': options.pop('cire-mincost-sops') or cls.CIRE_MINCOST_SOPS
+        o['cire-mincost'] = {
+            'invariants': oo.pop('cire-mincost-inv', cls.CIRE_MINCOST_INV),
+            'sops': oo.pop('cire-mincost-sops', cls.CIRE_MINCOST_SOPS)
         }
 
+        # GPU parallelism
+        o['par-collapse-ncores'] = 1  # Always use a collapse clause
+        o['par-collapse-work'] = 1  # Always use a collapse clause
+        o['par-chunk-nonaffine'] = oo.pop('par-chunk-nonaffine', cls.PAR_CHUNK_NONAFFINE)
+        o['par-dynamic-work'] = np.inf  # Always use static scheduling
+        o['par-nested'] = np.inf  # Never use nested parallelism
+
+        if oo:
+            raise InvalidOperator("Unsupported optimization options: [%s]"
+                                  % ", ".join(list(oo)))
+
+        kwargs['options'].update(o)
+
         return kwargs
 
     @classmethod
@@ -268,31 +278,28 @@ class DeviceOpenMPNoopOperator(OperatorCore):
     def _specialize_clusters(cls, clusters, **kwargs):
         options = kwargs['options']
         platform = kwargs['platform']
-
-        # To create temporaries
-        counter = generator()
-        template = lambda: "r%d" % counter()
+        sregistry = kwargs['sregistry']
 
         # Toposort+Fusion (the former to expose more fusion opportunities)
         clusters = fuse(clusters, toposort=True)
 
         # Hoist and optimize Dimension-invariant sub-expressions
-        clusters = cire(clusters, template, 'invariants', options, platform)
+        clusters = cire(clusters, 'invariants', sregistry, options, platform)
         clusters = Lift().process(clusters)
 
         # Reduce flops (potential arithmetic alterations)
-        clusters = extract_increments(clusters, template)
-        clusters = cire(clusters, template, 'sops', options, platform)
+        clusters = extract_increments(clusters, sregistry)
+        clusters = cire(clusters, 'sops', sregistry, options, platform)
         clusters = factorize(clusters)
         clusters = optimize_pows(clusters)
 
         # Reduce flops (no arithmetic alterations)
-        clusters = cse(clusters, template)
+        clusters = cse(clusters, sregistry)
 
         # Lifting may create fusion opportunities, which in turn may enable
         # further optimizations
         clusters = fuse(clusters)
-        clusters = eliminate_arrays(clusters, template)
+        clusters = eliminate_arrays(clusters)
 
         return clusters
 
@@ -300,16 +307,17 @@ class DeviceOpenMPNoopOperator(OperatorCore):
     @timed_pass(name='specializing.IET')
     def _specialize_iet(cls, graph, **kwargs):
         options = kwargs['options']
+        sregistry = kwargs['sregistry']
 
         # Distributed-memory parallelism
         if options['mpi']:
             mpiize(graph, mode=options['mpi'])
 
         # GPU parallelism via OpenMP offloading
-        DeviceOmpizer().make_parallel(graph)
+        DeviceOmpizer(sregistry, options).make_parallel(graph)
 
         # Symbol definitions
-        data_manager = DeviceOpenMPDataManager()
+        data_manager = DeviceOpenMPDataManager(sregistry)
         data_manager.place_ondevice(graph)
         data_manager.place_definitions(graph)
         data_manager.place_casts(graph)
@@ -323,6 +331,7 @@ class DeviceOpenMPOperator(DeviceOpenMPNoopOperator):
     @timed_pass(name='specializing.IET')
     def _specialize_iet(cls, graph, **kwargs):
         options = kwargs['options']
+        sregistry = kwargs['sregistry']
 
         # Distributed-memory parallelism
         optimize_halospots(graph)
@@ -330,13 +339,13 @@ class DeviceOpenMPOperator(DeviceOpenMPNoopOperator):
             mpiize(graph, mode=options['mpi'])
 
         # GPU parallelism via OpenMP offloading
-        DeviceOmpizer().make_parallel(graph)
+        DeviceOmpizer(sregistry, options).make_parallel(graph)
 
         # Misc optimizations
         hoist_prodders(graph)
 
         # Symbol definitions
-        data_manager = DeviceOpenMPDataManager()
+        data_manager = DeviceOpenMPDataManager(sregistry)
         data_manager.place_ondevice(graph)
         data_manager.place_definitions(graph)
         data_manager.place_casts(graph)
@@ -347,14 +356,15 @@ class DeviceOpenMPOperator(DeviceOpenMPNoopOperator):
 class DeviceOpenMPCustomOperator(DeviceOpenMPOperator):
 
     _known_passes = ('optcomms', 'openmp', 'mpi', 'prodders')
-    _known_passes_disabled = ('blocking', 'denormals', 'wrapping', 'simd')
+    _known_passes_disabled = ('blocking', 'denormals', 'simd')
     assert not (set(_known_passes) & set(_known_passes_disabled))
 
     @classmethod
     def _make_passes_mapper(cls, **kwargs):
         options = kwargs['options']
+        sregistry = kwargs['sregistry']
 
-        ompizer = DeviceOmpizer()
+        ompizer = DeviceOmpizer(sregistry, options)
 
         return {
             'optcomms': partial(optimize_halospots),
@@ -381,6 +391,7 @@ class DeviceOpenMPCustomOperator(DeviceOpenMPOperator):
     @timed_pass(name='specializing.IET')
     def _specialize_iet(cls, graph, **kwargs):
         options = kwargs['options']
+        sregistry = kwargs['sregistry']
         passes = as_tuple(kwargs['mode'])
 
         # Fetch passes to be called
@@ -402,7 +413,7 @@ class DeviceOpenMPCustomOperator(DeviceOpenMPOperator):
             passes_mapper['openmp'](graph)
 
         # Symbol definitions
-        data_manager = DeviceOpenMPDataManager()
+        data_manager = DeviceOpenMPDataManager(sregistry)
         data_manager.place_ondevice(graph)
         data_manager.place_definitions(graph)
         data_manager.place_casts(graph)
