@@ -5,6 +5,7 @@ import numpy as np
 
 from devito.data.allocators import ALLOC_FLAT
 from devito.data.utils import *
+from devito.logger import warning
 from devito.parameters import configuration
 from devito.tools import Tag, as_tuple, as_list, is_integer
 
@@ -76,6 +77,10 @@ class Data(np.ndarray):
         self._allocator.free(*self._memfree_args)
         self._memfree_args = None
 
+    def __reduce__(self):
+        warning("Pickling of `Data` objects is not supported. Casting to `numpy.ndarray`")
+        return np.array(self).__reduce__()
+
     def __array_finalize__(self, obj):
         # `self` is the newly created object
         # `obj` is the object from which `self` was created
