@@ -106,6 +106,14 @@ class Cluster(object):
     def itintervals(self):
         return self.ispace.itintervals
 
+    @property
+    def sub_iterators(self):
+        return self.ispace.sub_iterators
+
+    @property
+    def directions(self):
+        return self.ispace.directions
+
     @property
     def dspace(self):
         return self._dspace
