@@ -1,5 +1,4 @@
 from collections.abc import Iterable
-
 from operator import attrgetter
 
 from devito.symbolics import (retrieve_functions, retrieve_indexed, split_affine,
@@ -109,6 +108,9 @@ def lower_exprs(expressions, **kwargs):
             mapper[i] = f.indexed[indices]
 
         subs = kwargs.get('subs')
+        # Add dimensions map to the mapper in case dimensions are used
+        # as an expression, i.e. Eq(u, x, subdomain=xleft)
+        mapper.update(dimension_map)
         if subs:
             # Include the user-supplied substitutions, and use
             # `xreplace` for constant folding
