@@ -28,24 +28,31 @@ class Cached(object):
     @classmethod
     def _cache_key(cls, *args, **kwargs):
         """
-        A unique, deterministic cache key from the input arguments.
+        A unique, deterministic key from the input arguments.
 
         Notes
         -----
         To be implemented by subclasses.
+
+        Returns
+        -------
+        The cache key. It must be hashable.
         """
-        raise NotImplementedError
+        raise NotImplementedError("Subclass must implement _cache_key")
 
     @classmethod
     def _cache_get(cls, key):
         """
-        Retrieve the object corresponding to a given key. If the key is not in
-        the symbol cache or if the mapped object is not alive anymore, returns None.
+        Look up the cache for a given key.
 
         Parameters
         ----------
         key : object
             The cache key. It must be hashable.
+
+        Returns
+        -------
+        The object if in the cache and alive, otherwise None.
         """
         if key in _SymbolCache:
             # There is indeed an object mapped to `key`. But is it still alive?
@@ -59,7 +66,7 @@ class Cached(object):
         else:
             return None
 
-    def __init__(self, key):
+    def __init__(self, key, *aliases):
         """
         Store `self` in the symbol cache.
 
@@ -67,12 +74,16 @@ class Cached(object):
         ----------
         key : object
             The cache key. It must be hashable.
+        *aliases
+            Additional keys to which self is mapped.
         """
         # Precompute hash. This uniquely depends on the cache key
         self._cache_key_hash = hash(key)
 
         # Add ourselves to the symbol cache
-        _SymbolCache[key] = AugmentedWeakRef(self, self._cache_meta())
+        awr = AugmentedWeakRef(self, self._cache_meta())
+        for i in (key,) + aliases:
+            _SymbolCache[i] = awr
 
     def __init_cached__(self, key):
         """
@@ -84,13 +95,11 @@ class Cached(object):
             The cache key of the object whose state is used to initialize `self`.
             It must be hashable.
         """
-
-        self.__dict__ = _SymbolCache[key]().__dict__
+        self.__dict__ = _SymbolCache[key]().__dict__.copy()
 
     def __hash__(self):
         """
-        The hash value of an object that caches on its type is the
-        hash value of the type itself.
+        The hash value of a cached object is the hash of its cache key.
         """
         return self._cache_key_hash
 
