@@ -12,8 +12,8 @@ from devito.types.args import ArgProvider
 from devito.types.basic import Symbol, DataSymbol, Scalar
 
 __all__ = ['Dimension', 'SpaceDimension', 'TimeDimension', 'DefaultDimension',
-           'SteppingDimension', 'SubDimension', 'ConditionalDimension', 'dimensions',
-           'ModuloDimension', 'IncrDimension', 'ShiftedDimension']
+           'CustomDimension', 'SteppingDimension', 'SubDimension', 'ConditionalDimension',
+           'dimensions', 'ModuloDimension', 'IncrDimension', 'ShiftedDimension']
 
 
 Thickness = namedtuple('Thickness', 'left right')
@@ -95,6 +95,7 @@ class Dimension(ArgProvider):
     is_Time = False
 
     is_Default = False
+    is_Custom = False
     is_Derived = False
     is_NonlinearDerived = False
     is_Sub = False
@@ -166,6 +167,10 @@ class Dimension(ArgProvider):
     def max_name(self):
         return "%s_M" % self.name
 
+    @property
+    def is_const(self):
+        return False
+
     @property
     def root(self):
         return self
@@ -183,6 +188,10 @@ class Dimension(ArgProvider):
     def _defines(self):
         return frozenset({self})
 
+    @cached_property
+    def _defines_symbols(self):
+        return frozenset({self.symbolic_min, self.symbolic_max, self.symbolic_size})
+
     @property
     def _arg_names(self):
         """Tuple of argument names introduced by the Dimension."""
@@ -570,6 +579,10 @@ class SubDimension(DerivedDimension):
     def _thickness_map(self):
         return dict(self.thickness)
 
+    @cached_property
+    def _defines_symbols(self):
+        return super()._defines_symbols | frozenset(self._thickness_map)
+
     @cached_property
     def _offset_left(self):
         # The left extreme of the SubDimension can be related to either the
@@ -790,7 +803,7 @@ class SteppingDimension(DerivedDimension):
     def _arg_names(self):
         return (self.min_name, self.max_name, self.name) + self.parent._arg_names
 
-    def _arg_defaults(self, _min=None, size=None, **kwargs):
+    def _arg_defaults(self, _min=None, **kwargs):
         """
         A map of default argument values defined by this dimension.
 
@@ -798,14 +811,13 @@ class SteppingDimension(DerivedDimension):
         ----------
         _min : int, optional
             Minimum point as provided by data-carrying objects.
-        size : int, optional
-            Size as provided by data-carrying symbols.
 
         Notes
         -----
-        A SteppingDimension does not know its max point.
+        A SteppingDimension does not know its max point and therefore
+        does not have a size argument.
         """
-        return {self.parent.min_name: _min, self.size_name: size}
+        return {self.parent.min_name: _min}
 
     def _arg_values(self, *args, **kwargs):
         """
@@ -827,6 +839,9 @@ class SteppingDimension(DerivedDimension):
         return values
 
 
+# The Dimensions below are for internal use only
+
+
 class ModuloDimension(DerivedDimension):
 
     """
@@ -1060,6 +1075,36 @@ class ShiftedDimension(IncrDimension):
     _pickle_kwargs = []
 
 
+class CustomDimension(BasicDimension):
+
+    """
+    Dimension defining an iteration space with custom (known) size.
+
+    Notes
+    -----
+    This could be extended in the future for more customization (e.g., min point,
+    max point, etc.).
+    """
+
+    is_Custom = True
+
+    def __init_finalize__(self, name, symbolic_size=None):
+        self._symbolic_size = symbolic_size
+
+    @property
+    def symbolic_size(self):
+        return self._symbolic_size
+
+    def _arg_defaults(self):
+        return {}
+
+    def _arg_values(self, *args, **kwargs):
+        return {}
+
+    # Pickling support
+    _pickle_kwargs = ['symbolic_size']
+
+
 def dimensions(names):
     assert type(names) == str
     return tuple(Dimension(i) for i in names.split())
