@@ -133,7 +133,7 @@ class TensorFunction(AbstractTensor, Differentiable):
 
             def entry(i):
                 return sum(self[i, k]*other[k] for k in range(self.cols))
-            comps = [entry(i) for i in range(self.cols)]
+            comps = [entry(i) for i in range(self.rows)]
             func = vec_func(self, other)
             name = "%s%s" % (self.name, other.name)
             to = getattr(self, 'time_order', 0)
@@ -145,8 +145,9 @@ class TensorFunction(AbstractTensor, Differentiable):
 
             def entry(i, j):
                 return sum(self[i, k]*other[k, j] for k in range(self.cols))
-            comps = [[entry(i, j) for i in range(self.cols)]
-                     for j in range(self.rows)]
+            comps = [[entry(i, j) for j in range(self.cols)]
+                     for i in range(self.rows)]
+
             func = tens_func(self, other)
             name = "%s%s" % (self.name, other.name)
             to = getattr(self, 'time_order', 0)
