@@ -35,14 +35,16 @@ class AcousticWaveSolver(object):
         self.space_order = space_order
         self.kernel = kernel
 
-        # Time step can be \sqrt{3}=1.73 bigger with 4th order
-        self.dt = self.model.critical_dt
-        if self.kernel == 'OT4':
-            self.dt = model.dtype(self.dt*1.73)
-
         # Cache compiler options
         self._kwargs = kwargs
 
+    @property
+    def dt(self):
+        # Time step can be \sqrt{3}=1.73 bigger with 4th order
+        if self.kernel == 'OT4':
+            return self.model.dtype(1.73 * self.model.critical_dt)
+        return self.model.critical_dt
+
     @memoized_meth
     def op_fwd(self, save=None):
         """Cached operator for forward runs with buffered wavefield"""
