@@ -1,6 +1,5 @@
 import numpy as np
-from sympy import sin, Abs, finite_diff_weights
-
+from sympy import sin, Abs, finite_diff_weights as fd_w
 
 from devito import (Grid, SubDomain, Function, Constant,
                     SubDimension, Eq, Inc, Operator, div)
@@ -11,9 +10,9 @@ __all__ = ['SeismicModel', 'Model', 'ModelElastic',
            'ModelViscoelastic', 'ModelViscoacoustic']
 
 
-def initialize_damp(damp, nbl, spacing, abc_type="damp"):
+def initialize_damp(damp, nbl, spacing, abc_type="damp", fs=False):
     """
-    Initialise damping field with an absorbing boundary layer.
+    Initialize damping field with an absorbing boundary layer.
 
     Parameters
     ----------
@@ -32,13 +31,14 @@ def initialize_damp(damp, nbl, spacing, abc_type="damp"):
 
     eqs = [Eq(damp, 1.0)] if abc_type == "mask" else []
     for d in damp.dimensions:
-        # left
-        dim_l = SubDimension.left(name='abc_%s_l' % d.name, parent=d,
-                                  thickness=nbl)
-        pos = Abs((nbl - (dim_l - d.symbolic_min) + 1) / float(nbl))
-        val = dampcoeff * (pos - sin(2*np.pi*pos)/(2*np.pi))
-        val = -val if abc_type == "mask" else val
-        eqs += [Inc(damp.subs({d: dim_l}), val/d.spacing)]
+        if not fs or d is not damp.dimensions[-1]:
+            # left
+            dim_l = SubDimension.left(name='abc_%s_l' % d.name, parent=d,
+                                      thickness=nbl)
+            pos = Abs((nbl - (dim_l - d.symbolic_min) + 1) / float(nbl))
+            val = dampcoeff * (pos - sin(2*np.pi*pos)/(2*np.pi))
+            val = -val if abc_type == "mask" else val
+            eqs += [Inc(damp.subs({d: dim_l}), val/d.spacing)]
         # right
         dim_r = SubDimension.right(name='abc_%s_r' % d.name, parent=d,
                                    thickness=nbl)
@@ -52,14 +52,35 @@ def initialize_damp(damp, nbl, spacing, abc_type="damp"):
 
 class PhysicalDomain(SubDomain):
 
-    name = 'phydomain'
+    name = 'physdomain'
 
-    def __init__(self, nbl):
+    def __init__(self, so, fs=False):
         super(PhysicalDomain, self).__init__()
-        self.nbl = nbl
+        self.so = so
+        self.fs = fs
 
     def define(self, dimensions):
-        return {d: ('middle', self.nbl, self.nbl) for d in dimensions}
+        map_d = {d: d for d in dimensions}
+        if self.fs:
+            map_d[dimensions[-1]] = ('middle', self.so, 0)
+        return map_d
+
+
+class FSDomain(SubDomain):
+
+    name = 'fsdomain'
+
+    def __init__(self, so):
+        super(FSDomain, self).__init__()
+        self.size = so
+
+    def define(self, dimensions):
+        """
+        Definition of the upper section of the domain for wrapped indices FS.
+        """
+
+        return {d: (d if not d == dimensions[-1] else ('left', self.size))
+                for d in dimensions}
 
 
 class GenericModel(object):
@@ -67,19 +88,29 @@ class GenericModel(object):
     General model class with common properties
     """
     def __init__(self, origin, spacing, shape, space_order, nbl=20,
-                 dtype=np.float32, subdomains=(), bcs="damp", grid=None):
+                 dtype=np.float32, subdomains=(), bcs="damp", grid=None,
+                 fs=False):
         self.shape = shape
         self.space_order = space_order
         self.nbl = int(nbl)
         self.origin = tuple([dtype(o) for o in origin])
+        self.fs = fs
+        # Default setup
+        origin_pml = [dtype(o - s*nbl) for o, s in zip(origin, spacing)]
+        shape_pml = np.array(shape) + 2 * self.nbl
+
+        # Model size depending on freesurface
+        physdomain = PhysicalDomain(space_order, fs=fs)
+        subdomains = subdomains + (physdomain,)
+        if fs:
+            fsdomain = FSDomain(space_order)
+            subdomains = subdomains + (fsdomain,)
+            origin_pml[-1] = origin[-1]
+            shape_pml[-1] -= self.nbl
 
         # Origin of the computational domain with boundary to inject/interpolate
         # at the correct index
         if grid is None:
-            origin_pml = tuple([dtype(o - s*nbl) for o, s in zip(origin, spacing)])
-            phydomain = PhysicalDomain(self.nbl)
-            subdomains = subdomains + (phydomain, )
-            shape_pml = np.array(shape) + 2 * self.nbl
             # Physical extent is calculated per cell, so shape - 1
             extent = tuple(np.array(spacing) * (shape_pml - 1))
             self.grid = Grid(extent=extent, shape=shape_pml, origin=origin_pml,
@@ -87,18 +118,27 @@ class GenericModel(object):
         else:
             self.grid = grid
 
+        # Create dampening field as symbol `damp`
         if self.nbl != 0:
-            # Create dampening field as symbol `damp`
             self.damp = Function(name="damp", grid=self.grid)
             if callable(bcs):
                 bcs(self.damp, self.nbl)
             else:
-                initialize_damp(self.damp, self.nbl, self.spacing, abc_type=bcs)
+                initialize_damp(self.damp, self.nbl, self.spacing, abc_type=bcs, fs=fs)
             self._physical_parameters = ['damp']
         else:
             self.damp = 1 if bcs == "mask" else 0
             self._physical_parameters = []
 
+    @property
+    def padsizes(self):
+        """
+        Padding size for each dimension.
+        """
+        padsizes = [(self.nbl, self.nbl) for _ in range(self.dim-1)]
+        padsizes.append((0 if self.fs else self.nbl, self.nbl))
+        return padsizes
+
     def physical_params(self, **kwargs):
         """
         Return all set physical parameters and update to input values if provided
@@ -113,7 +153,7 @@ class GenericModel(object):
         if isinstance(field, np.ndarray):
             function = Function(name=name, grid=self.grid, space_order=space_order,
                                 parameter=is_param)
-            initialize_function(function, field, self.nbl)
+            initialize_function(function, field, self.padsizes)
         else:
             function = Constant(name=name, value=field, dtype=self.grid.dtype)
         self._physical_parameters.append(name)
@@ -208,16 +248,21 @@ class SeismicModel(GenericModel):
     _known_parameters = ['vp', 'damp', 'vs', 'b', 'epsilon', 'delta',
                          'theta', 'phi', 'qp', 'qs', 'lam', 'mu']
 
-    def __init__(self, origin, spacing, shape, space_order, vp, nbl=20,
+    def __init__(self, origin, spacing, shape, space_order, vp, nbl=20, fs=False,
                  dtype=np.float32, subdomains=(), bcs="mask", grid=None, **kwargs):
         super(SeismicModel, self).__init__(origin, spacing, shape, space_order, nbl,
-                                           dtype, subdomains, grid=grid, bcs=bcs)
+                                           dtype, subdomains, grid=grid, bcs=bcs, fs=fs)
 
         # Initialize physics
         self._initialize_physics(vp, space_order, **kwargs)
 
         # User provided dt
         self._dt = kwargs.get('dt')
+        # Some wave equation need a rescaled dt that can't be infered from the model
+        # parameters, such as isoacoustic OT4 that can use a dt sqrt(3) larger than
+        # isoacoustic OT2. This property should be set from a wavesolver or after model
+        # instanciation only via model.dt_scale = value.
+        self._dt_scale = 1
 
     def _initialize_physics(self, vp, space_order, **kwargs):
         """
@@ -258,31 +303,36 @@ class SeismicModel(GenericModel):
             return np.sqrt(mmin(self.b) * (mmax(self.lam) + 2 * mmax(self.mu)))
 
     @property
-    def _scale(self):
+    def _thomsen_scale(self):
         # Update scale for tti
         if 'epsilon' in self._physical_parameters:
             return np.sqrt(1 + 2 * mmax(self.epsilon))
         return 1
 
+    @property
+    def dt_scale(self):
+        return self._dt_scale
+
+    @dt_scale.setter
+    def dt_scale(self, val):
+        self._dt_scale = val
+
     @property
     def _cfl_coeff(self):
         """
-        Courant number from the physics.
-        For the elastic case, we use the general `.85/sqrt(ndim)`.
-
-        For te acoustic case, we can get an accurate estimate from the spatial order.
-        One dan show that C = sqrt(a1/a2) where:
-        - a1 is the sum of absolute value of FD coefficients in time
-        - a2 is the sum of absolute value of FD coefficients in space
-        https://library.seg.org/doi/pdf/10.1190/1.1444605
+        Courant number from the physics and spatial discretization order.
+        The CFL coefficients are described in:
+        - https://doi.org/10.1137/0916052 for the elastic case
+        - https://library.seg.org/doi/pdf/10.1190/1.1444605 for the acoustic case
         """
         # Elasic coefficient (see e.g )
         if 'lam' in self._physical_parameters or 'vs' in self._physical_parameters:
-            return .85 / np.sqrt(self.grid.dim)
+            coeffs = fd_w(1, range(-self.space_order//2+1, self.space_order//2+1), .5)
+            c_fd = sum(np.abs(coeffs[-1][-1])) / 2
+            return np.sqrt(self.dim) / self.dim / c_fd
         a1 = 4  # 2nd order in time
-        coeffs = finite_diff_weights(2, range(-self.space_order, self.space_order+1), 0)
-        a2 = float(self.grid.dim * sum(np.abs(coeffs[-1][-1])))
-        return np.sqrt(a1/a2)
+        coeffs = fd_w(2, range(-self.space_order, self.space_order+1), 0)[-1][-1]
+        return np.sqrt(a1/float(self.grid.dim * sum(np.abs(coeffs))))
 
     @property
     def critical_dt(self):
@@ -293,7 +343,8 @@ class SeismicModel(GenericModel):
         #
         # The CFL condtion is then given by
         # dt <= coeff * h / (max(velocity))
-        dt = self._cfl_coeff * np.min(self.spacing) / (self._scale*self._max_vp)
+        dt = self._cfl_coeff * np.min(self.spacing) / (self._thomsen_scale*self._max_vp)
+        dt = self.dt_scale * dt
         if self._dt:
             assert self._dt < dt
             return self._dt
