@@ -39,11 +39,14 @@ class AnisotropicWaveSolver(object):
                     "but got %s" % space_order)
 
         self.space_order = space_order
-        self.dt = self.model.critical_dt
 
         # Cache compiler options
         self._kwargs = kwargs
 
+    @property
+    def dt(self):
+        return self.model.critical_dt
+
     @memoized_meth
     def op_fwd(self, kernel='centered', save=False):
         """Cached operator for forward runs with buffered wavefield"""
@@ -141,7 +144,7 @@ class AnisotropicWaveSolver(object):
 
     def adjoint(self, rec, srca=None, p=None, r=None, vp=None,
                 epsilon=None, delta=None, theta=None, phi=None,
-                save=None, **kwargs):
+                save=None, kernel='centered', **kwargs):
         """
         Adjoint modelling function that creates the necessary
         data objects for running an adjoint modelling operator.
@@ -170,6 +173,8 @@ class AnisotropicWaveSolver(object):
         -------
         Adjoint source, wavefield and performance summary.
         """
+        if kernel != 'centered':
+            raise RuntimeError('Only centered kernel is supported for the adjoint')
 
         time_order = 2
         stagg_p = stagg_r = None
